"""
Service class
------------------
Service object for interfacing with the Project service API

"""

import logging

from ...client import Client
from ...services import BaseService

from .service import Service
